/*****************************************************************************
 *  Copyright 2005 ELCOM, a.s.  All Rights Reserved.                         *
 *****************************************************************************/
/*****************************************************************************
 *  GM SSP Konstanter (VISA)
 *****************************************************************************/



#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <locale.h>  
#include <visa.h>

#include "gmpsp_utility.h"
#include "gmpsp.h"
/*= Macros ==================================================================*/

#define CHECKERR_CU(fCal) if (gmpsp_status = (fCal), gmpsp_status < VI_SUCCESS) \
    return gmpsp_initCleanUp (rmSession, instrSession, rssmb_status); else

/*= DEFINES =================================================================*/
#define GMPSP_REVISION          "Rev 1.0, 09/2007, CVI 8.0" /* Instrument driver revision */
#define BUFFER_SIZE             2048L    /* File I/O buffer size */

#define GMPSP_DEFAULT_TIMEOUT   10000   /* Default Timeout value in ms (10 sec) */

/*****************************************************************************/
/*= INSTRUMENT-DEPENDENT COMMAND ARRAYS =====================================*/
/*****************************************************************************/
static ViString limitArr[] = {"_L","_H"};
static ViString outputModeArr[] = {"OFF","CV","CC","OL","CP",VI_NULL};
static ViString triggerModeArr[] = {"OFF","OUT","SQS","SEQ","LLO","MIN","AIX","AIU","AII",VI_NULL};
static ViString signalArr[] = {"OFF","ON","OUT","MODE","SEQ","SSET","U_LO","U_HI","I_LO","I_HI",VI_NULL};
static ViString analogArr[] = {"OFF","ON","SSET",VI_NULL};
static ViString powerOnArr[] = {"RST","RCL","SBY",VI_NULL};
static ViString displayArr[] = {"OFF","UO","US","PS",VI_NULL};
static ViString displayBArr[] = {"OFF","IO","IS","PO",VI_NULL};
static ViString regulationArr[] = {"R","L",VI_NULL};
static ViString funcArr[] = {"CLR","NF","RU","RI","SOFF","S_ON","AUOF","AUON","AUSS","AIOF","AION","AISS",
                             "R01","R02","R03","R04","R05","R06","R07","R08","R09","R10","R11","R12",
                             "S01","S02","S03","S04","S05","S06","S07","S08","S09","S10","S11","S12",
                             VI_NULL};
static ViString triggerArr[] = {"GO","STRT","STOP","STEP","HOLD",VI_NULL};
static ViString triggerStatusArr[] = {"RDY","RUN","HALT",VI_NULL};
static ViString registerEnableArr[] = {"ERAE","ERBE","ERCE","ERA","ERB","ERC","CRA","CRB"};

/*****************************************************************************/
/*= UTILITY ROUTINE DECLARATIONS (Non-Exportable Functions) =================*/
/*****************************************************************************/
ViStatus gmpsp_initCleanUp (ViSession openRMSession, ViPSession openInstrSession,
                             ViStatus currentStatus);
ViStatus gmpsp_defaultInstrSetup (ViSession openInstrSession);

/*****************************************************************************
 * Function: Initialize
 * Purpose:  This function performs the following initialization actions:
 *           
 *           - Opens a session to the Default Resource Manager resource and a 
 *           session to the specified device using the interface and address 
 *           specified in the Resource_Name control.
 *           
 *           - Performs an identification query on the Instrument.
 *           
 *           - Resets the instrument to a known state.
 *           
 *           - Sends initialization commands to the instrument that set any 
 *           necessary programmatic variables such as Headers Off, Short 
 *           Command form, and Data Transfer Binary to the state necessary for 
 *           the operation of the instrument driver.
 *           
 *           - Returns an Instrument Handle which is used to differentiate 
 *           between different sessions of this instrument driver.
 *           
 *           - Each time this function is invoked a Unique Session is opened.  
 *           It is possible to have more than one session open for the same 
 *           resource.
 *           
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_init(
    ViRsrc  resourceName,
    ViBoolean   idQuery,
    ViBoolean   resetDevice,
    ViSession*  instrSession
)
{
    ViStatus            gmpsp_status = VI_SUCCESS;
    ViSession           rmSession = 0;
    ViChar              buffer[BUFFER_SIZE] = "";
    gmpsp_instrPtr      instrPtr = NULL;

    /* --- Set standard "C" locale --- */
    {
       char *locale;

       locale = setlocale(LC_ALL, "C");
    }
    
    /*- Check input parameter ranges ----------------------------------------*/
    CHECKPARAM (gmpsp_invalidViBooleanRange (idQuery), 2);
    CHECKPARAM (gmpsp_invalidViBooleanRange (resetDevice), 3);

    /*- Open instrument session ---------------------------------------------*/
    CHECKERR( viOpenDefaultRM (&rmSession));

    if ((gmpsp_status = viOpen (rmSession, resourceName, VI_LOAD_CONFIG,
                                 VI_NULL, instrSession)) < 0)
    {
        (void)viClose (rmSession);
        return gmpsp_status;
    }

    /* Determine if the structure has been initialized for the current VISA  */
    /* Session and malloc if it has not.                                     */

    
    if (instrPtr == NULL)
        instrPtr = calloc (1, sizeof (struct gmpsp_statusDataRanges));
        

    CHECKERR( viSetAttribute (*instrSession, VI_ATTR_USER_DATA, (ViUInt32)instrPtr));

    /*- Configure VISA Formatted I/O ----------------------------------------*/
    CHECKERR( viSetAttribute (*instrSession, VI_ATTR_TMO_VALUE,
                                 GMPSP_DEFAULT_TIMEOUT));
    CHECKERR( viSetBuf (*instrSession, (ViUInt16) (VI_READ_BUF|VI_WRITE_BUF), 4000));
    CHECKERR( viSetAttribute (*instrSession, VI_ATTR_WR_BUF_OPER_MODE,
                                 VI_FLUSH_ON_ACCESS));
    CHECKERR( viSetAttribute (*instrSession, VI_ATTR_RD_BUF_OPER_MODE,
                                 VI_FLUSH_ON_ACCESS));

    /* -------------------------------------------------------------------- */
    /* Drawback of Agilent VISA is patched by adding extra termination      */
    /* character "\n" to each I/O operation                                 */
    /*                                                                      */
    /* VI_ATTR_SEND_END_EN will stay with this attribute set to VI_TRUE     */
    /* -------------------------------------------------------------------- */

    CHECKERR( viSetAttribute (*instrSession, VI_ATTR_SEND_END_EN, VI_TRUE));

    // read/reset ESR which may still indicate device dependent error from 
    // previous handling!
    (void) gmpsp_checkStatus(*instrSession);
    
    /*- Identification Query ------------------------------------------------*/
    if (idQuery == VI_TRUE)
    {
        CHECKERR( viQueryf (*instrSession, "*IDN?\n", "%*[^ ] %[^,],%*[\r\n]",
                               buffer));
        /* Check if instrument is supported or not */
        if (strncmp (buffer, "GOSSEN-METRAWATT", 16) != 0)
            return gmpsp_initCleanUp (rmSession, instrSession, VI_ERROR_FAIL_ID_QUERY);
    }

    /*- Reset instrument ----------------------------------------------------*/
    if (resetDevice == VI_TRUE)
        CHECKERR( gmpsp_reset (*instrSession));
    else  /*- Send Default Instrument Setup ---------------------------------*/
        CHECKERR( gmpsp_defaultInstrSetup (*instrSession));

    return gmpsp_status;
}

/*****************************************************************************
 * Function: Configure Voltage
 * Purpose:  This function configures voltage parameters.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_ConfigureVoltage(
    ViSession   instrSession,
    ViReal64    setVoltage,
    ViReal64    voltageLimitLow,
    ViReal64    voltageLimitHigh
)
{
    ViStatus gmpsp_status = 0;
    
    CHECKERR (gmpsp_SetVoltageLimitValue (instrSession, GMPSP_MAX, 60.0));
    CHECKERR (gmpsp_SetVoltageLimitValue (instrSession, GMPSP_MIN, 0.0));
    CHECKERR (gmpsp_SetVoltageValue (instrSession, voltageLimitLow));

    CHECKERR (gmpsp_SetVoltageLimitValue (instrSession, GMPSP_MIN, voltageLimitLow));
    
    gmpsp_status = gmpsp_SetVoltageLimitValue (instrSession, GMPSP_MAX, voltageLimitHigh);
    REMAP (3,4);
    
    return gmpsp_SetVoltageValue (instrSession, setVoltage);
}

/*****************************************************************************
 * Function: Configure Overvoltage Protection
 * Purpose:  This function configures overvoltage protection parameters.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_ConfigureOvervoltageProtection(
    ViSession   instrSession,
    ViBoolean   ovpState,
    ViReal64    ovpLevel,
    ViReal64    ovpDelay
)
{
    ViStatus gmpsp_status = 0;
    
    gmpsp_status = gmpsp_SetOvervoltageProtectionValue (instrSession, ovpLevel);
    REMAP (2,3);
    
    gmpsp_status = gmpsp_SetOvervoltageProtectionDelay (instrSession, ovpDelay);
    REMAP (2,4);
    
    return gmpsp_SetOvervoltageProtectionState (instrSession, ovpState);
}

/*****************************************************************************
 * Function: Set Voltage Value
 * Purpose:  This function sets the desired Voltage Value.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_SetVoltageValue(
    ViSession   instrSession,
    ViReal64    setVoltage
)
{
    return gmpsp_setViReal64 (instrSession,
                              "USET",
                              setVoltage,
                              0.0, 60.0,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Voltage Value
 * Purpose:  This function queries the desired Voltage Value.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_GetVoltageValue(
    ViSession   instrSession,
    ViReal64*   setVoltage
)
{
    return gmpsp_getViReal64 (instrSession,
                              "USET?",
                              setVoltage,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set Voltage Limit Value
 * Purpose:  This function sets the Voltage Limit Value.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_SetVoltageLimitValue(
    ViSession   instrSession,
    ViInt32 limitType,
    ViReal64    voltageValue
)
{
    ViChar  buffer[BUFFER_SIZE] = "";
    
    CHECKPARAM (gmpsp_invalidViInt32Range (limitType, 0, 1), 2);
    
    sprintf (buffer, "UL%s", limitArr[limitType]);

    return gmpsp_setViReal64 (instrSession,
                              buffer,
                              voltageValue,
                              0.0, 60.0,
                              VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Get Voltage Limit Value
 * Purpose:  This function queries selected Voltage Limit Value.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_GetVoltageLimitValue(
    ViSession   instrSession,
    ViInt32 limitType,
    ViReal64*   voltageValue
)
{
    ViChar  buffer[BUFFER_SIZE] = "";
    
    CHECKPARAM (gmpsp_invalidViInt32Range (limitType, 0, 1), 2);
    
    sprintf (buffer, "UL%s?", limitArr[limitType]);

    return gmpsp_getViReal64 (instrSession,
                              buffer,
                              voltageValue,
                              VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Set Overvoltage Protection State
 * Purpose:  This function sets the overvoltage protection state.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_SetOvervoltageProtectionState(
    ViSession   instrSession,
    ViBoolean   ovpState
)
{
    return gmpsp_setViBoolean (instrSession,
                               "OVP",
                               ovpState,
                               VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Overvoltage Protection State
 * Purpose:  This function queries the overvoltage protection state.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_GetOvervoltageProtectionState(
    ViSession   instrSession,
    ViBoolean*  ovpState
)
{
    return gmpsp_getViBoolean (instrSession,
                               "OVP?",
                               ovpState,
                               VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set Overvoltage Protection Value
 * Purpose:  This function sets the overvoltage protection value.
 *           
 *           Remote-control command(s):
 *           OVSET
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_SetOvervoltageProtectionValue(
    ViSession   instrSession,
    ViReal64    ovpLevel
)
{
    return gmpsp_setViReal64 (instrSession,
                              "OVSET",
                              ovpLevel,
                              3.0, 80.0,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Overvoltage Protection Value
 * Purpose:  This function queries the overvoltage protection value.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_GetOvervoltageProtectionValue(
    ViSession   instrSession,
    ViReal64*   ovpLevel
)
{
    return gmpsp_getViReal64 (instrSession,
                              "OVSET?",
                              ovpLevel,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set Overvoltage Protection Delay
 * Purpose:  This function sets the overvoltage protection delay.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_SetOvervoltageProtectionDelay(
    ViSession   instrSession,
    ViReal64    ovpDelay
)
{
    return gmpsp_setViReal64 (instrSession,
                              "OV_DELAY",
                              ovpDelay,
                              0.0, 65.535,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Overvoltage Protection Delay
 * Purpose:  This function queries the overvoltage protection delay.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_GetOvervoltageProtectionDelay(
    ViSession   instrSession,
    ViReal64*   ovpDelay
)
{
    return gmpsp_getViReal64 (instrSession,
                              "OV_DELAY?",
                              ovpDelay,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Configure Current
 * Purpose:  This function configures current parameters.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_ConfigureCurrent(
    ViSession   instrSession,
    ViReal64    setCurrent,
    ViReal64    currentLimitLow,
    ViReal64    currentLimitHigh
)
{
    ViStatus gmpsp_status = 0;
    
    CHECKERR (gmpsp_SetCurrentLimitValue (instrSession, GMPSP_MAX, 60.0));
    CHECKERR (gmpsp_SetCurrentLimitValue (instrSession, GMPSP_MIN, 0.0));
    CHECKERR (gmpsp_SetCurrentValue (instrSession, currentLimitLow));

    CHECKERR (gmpsp_SetCurrentLimitValue (instrSession, GMPSP_MIN, currentLimitLow));
    
    gmpsp_status = gmpsp_SetCurrentLimitValue (instrSession, GMPSP_MAX, currentLimitHigh);
    REMAP (3,4);
    
    return gmpsp_SetCurrentValue (instrSession, setCurrent);
}

/*****************************************************************************
 * Function: Configure Overcurrent Protection
 * Purpose:  This function configures overcurrent protection parameters.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_ConfigureOvercurrentProtection(
    ViSession   instrSession,
    ViBoolean   ocpState,
    ViReal64    ocpLevel,
    ViReal64    ocpDelay
)
{
    ViStatus gmpsp_status = 0;
    
    gmpsp_status = gmpsp_SetOvercurrentProtectionValue (instrSession, ocpLevel);
    REMAP (2,3);
    
    gmpsp_status = gmpsp_SetOvercurrentProtectionDelay (instrSession, ocpDelay);
    REMAP (2,4);
    
    return gmpsp_SetOvercurrentProtectionState (instrSession, ocpState);
}

/*****************************************************************************
 * Function: Set Current Value
 * Purpose:  This function sets the desired current value.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_SetCurrentValue(
    ViSession   instrSession,
    ViReal64    setCurrent
)
{
    return gmpsp_setViReal64 (instrSession,
                              "ISET",
                              setCurrent,
                              0.0, 180.0,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Current Value
 * Purpose:  This function queries the desired current value.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_GetCurrentValue(
    ViSession   instrSession,
    ViReal64*   setCurrent
)
{
    return gmpsp_getViReal64 (instrSession,
                              "ISET?",
                              setCurrent,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set Current Limit Value
 * Purpose:  This function sets the current limit value.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_SetCurrentLimitValue(
    ViSession   instrSession,
    ViInt32 limitType,
    ViReal64    currentValue
)
{
    ViChar  buffer[BUFFER_SIZE] = "";
    
    CHECKPARAM (gmpsp_invalidViInt32Range (limitType, 0, 1), 2);
    
    sprintf (buffer, "IL%s", limitArr[limitType]);

    return gmpsp_setViReal64 (instrSession,
                              buffer,
                              currentValue,
                              0.0, 180.0,
                              VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Get Current Limit Value
 * Purpose:  This function queries the current limit value.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_GetCurrentLimitValue(
    ViSession   instrSession,
    ViInt32 limitType,
    ViReal64*   currentValue
)
{
    ViChar  buffer[BUFFER_SIZE] = "";
    
    CHECKPARAM (gmpsp_invalidViInt32Range (limitType, 0, 1), 2);
    
    sprintf (buffer, "IL%s?", limitArr[limitType]);

    return gmpsp_getViReal64 (instrSession,
                              buffer,
                              currentValue,
                              VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Set Overcurrent Protection State
 * Purpose:  This function sets the overcurrent protection state.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_SetOvercurrentProtectionState(
    ViSession   instrSession,
    ViBoolean   ocpState
)
{
    return gmpsp_setViBoolean (instrSession,
                               "OCP",
                               ocpState,
                               VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Overcurrent Protection State
 * Purpose:  This function queries the overcurrent protection state.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_GetOvercurrentProtectionState(
    ViSession   instrSession,
    ViBoolean*  ocpState
)
{
    return gmpsp_getViBoolean (instrSession,
                               "OCP?",
                               ocpState,
                               VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set Overcurrent Protection Value
 * Purpose:  This function sets the overcurrent protection value.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_SetOvercurrentProtectionValue(
    ViSession   instrSession,
    ViReal64    ocpLevel
)
{
    return gmpsp_setViReal64 (instrSession,
                              "OCSET",
                              ocpLevel,
                              3.0, 240.0,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Overcurrent Protection Value
 * Purpose:  This function queries the overcurrent protection value.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_GetOvercurrentProtectionValue(
    ViSession   instrSession,
    ViReal64*   ocpLevel
)
{
    return gmpsp_getViReal64 (instrSession,
                              "OCSET?",
                              ocpLevel,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set Overcurrent Protection Delay
 * Purpose:  This function sets the overcurrent protection delay.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_SetOvercurrentProtectionDelay(
    ViSession   instrSession,
    ViReal64    ocpDelay
)
{
    return gmpsp_setViReal64 (instrSession,
                              "OC_DELAY",
                              ocpDelay,
                              0.0, 65.535,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Overcurrent Protection Delay
 * Purpose:  This function queries the overcurrent protection delay.
 *           
 *           Remote-control command(s):
 *           OC_DELAY?
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_GetOvercurrentProtectionDelay(
    ViSession   instrSession,
    ViReal64*   ocpDelay
)
{
    return gmpsp_getViReal64 (instrSession,
                              "OC_DELAY?",
                              ocpDelay,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set Power Value
 * Purpose:  This function sets the desired power value.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_SetPowerValue(
    ViSession   instrSession,
    ViReal64    setPower
)
{
    return gmpsp_setViReal64 (instrSession,
                              "PSET",
                              setPower,
                              0.0, 4000.0,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Power Value
 * Purpose:  This function queries the desired power value.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_GetPowerValue(
    ViSession   instrSession,
    ViReal64*   setPower
)
{
    return gmpsp_getViReal64 (instrSession,
                              "PSET?",
                              setPower,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Configure Output State
 * Purpose:  This function selects power output ON/OFF.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_ConfigureOutputState(
    ViSession   instrSession,
    ViBoolean   outputState
)
{
    return gmpsp_setViBoolean (instrSession,
                               "OUTPUT",
                               outputState,
                               VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Output State
 * Purpose:  This function queries power output state.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_GetOutputState(
    ViSession   instrSession,
    ViBoolean*  outputState
)
{
    return gmpsp_getViBoolean (instrSession,
                               "OUTPUT?",
                               outputState,
                               VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Output Mode
 * Purpose:  This function queries the current operating mode of the output.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_GetOutputMode(
    ViSession   instrSession,
    ViInt32*    outputMode
)
{
    return gmpsp_getEnum (instrSession,
                          "MODE?",
                          outputMode,
                          outputModeArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Configure Trigger Mode
 * Purpose:  This function defines Trigger Modes. The trigger mode determines 
 *           the effect of the floating optocoupler input Trigger on the Analog 
 *           Interface.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_ConfigureTriggerMode(
    ViSession   instrSession,
    ViInt32 triggerMode1,
    ViInt32 triggerMode2
)
{
    ViChar  buffer[BUFFER_SIZE] = "";
    
    CHECKPARAM (gmpsp_invalidViInt32Range (triggerMode1, 0, 8), 2);
    CHECKPARAM (gmpsp_invalidViInt32Range (triggerMode2, 0, 8), 3);
    
    sprintf (buffer, "T_MODE %s,%s", triggerModeArr[triggerMode1], triggerModeArr[triggerMode2]);

    return gmpsp_sendCmd (instrSession, buffer);
}

/*****************************************************************************
 * Function: Configure Signal Function
 * Purpose:  This function defines specific function for Signal1 to Signal3 
 *           outputs on Analog Interface. 
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_ConfigureSignalFunction(
    ViSession   instrSession,
    ViInt32 signal1,
    ViInt32 signal2,
    ViInt32 signal3
)
{
    ViChar  buffer[BUFFER_SIZE] = "";
    
    CHECKPARAM (gmpsp_invalidViInt32Range (signal1, 0, 9), 2);
    CHECKPARAM (gmpsp_invalidViInt32Range (signal2, 0, 9), 3);
    CHECKPARAM (gmpsp_invalidViInt32Range (signal3, 0, 9), 4);
    
    sprintf (buffer, "SIG123 %s,%s,%s", signalArr[signal1], signalArr[signal2], signalArr[signal3]);

    return gmpsp_sendCmd (instrSession, buffer);
}

/*****************************************************************************
 * Function: Configure Signal Output State
 * Purpose:  This function configures analog interface states.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_ConfigureSignalOutputState(
    ViSession   instrSession,
    ViInt32 uext,
    ViInt32 iext,
    ViBoolean   sset
)
{
    ViStatus gmpsp_status = 0;
    
    gmpsp_status = gmpsp_SetSset (instrSession, sset);
    REMAP (2,4);
    return gmpsp_SetAnalogInterfaceState (instrSession, uext, iext);
}

/*****************************************************************************
 * Function: Get Trigger Mode
 * Purpose:  This function queries Trigger Modes.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_GetTriggerMode(
    ViSession   instrSession,
    ViInt32*    triggerMode1,
    ViInt32*    triggerMode2
)
{
    ViStatus gmpsp_status = 0;
    ViChar   buffer[BUFFER_SIZE];
    ViChar   *p2buf;
    
    CHECKERR (gmpsp_getViString (instrSession, "T_MODE?", buffer, BUFFER_SIZE, 0));
    p2buf = strtok (buffer, ",");
    
    CHECKERR (gmpsp_getIndexFromString (p2buf, triggerModeArr, triggerMode1));
    p2buf = strtok (NULL, ",");
    CHECKERR (gmpsp_getIndexFromString (p2buf, triggerModeArr, triggerMode2));
    
    return gmpsp_status;
}

/*****************************************************************************
 * Function: Get Signal Function
 * Purpose:  This function queries specific function for Signal1 to Signal3 
 *           outputs on Analog Interface. 
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_GetSignalFunction(
    ViSession   instrSession,
    ViInt32*    signal1,
    ViInt32*    signal2,
    ViInt32*    signal3
)
{
    ViStatus gmpsp_status = 0;
    ViChar   buffer[BUFFER_SIZE];
    ViChar   *p2buf;
    
    CHECKERR (gmpsp_getViString (instrSession, "SIG123?", buffer, BUFFER_SIZE, 0));
    
    p2buf = strtok (buffer, ",");
    while (*p2buf == ' ') p2buf++;
    
    CHECKERR (gmpsp_getIndexFromString (p2buf, signalArr, signal1));
    p2buf = strtok (NULL, ",");
    while (*p2buf == ' ') p2buf++;
    CHECKERR (gmpsp_getIndexFromString (p2buf, signalArr, signal2));
    p2buf = strtok (NULL, ",");
    while (*p2buf == ' ') p2buf++;
    CHECKERR (gmpsp_getIndexFromString (p2buf, signalArr, signal3));
    
    return gmpsp_status;
}

/*****************************************************************************
 * Function: Set Sset
 * Purpose:  This function sets Signal Set (SSet) functionality in the 
 *           instrument.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_SetSset(
    ViSession   instrSession,
    ViBoolean   sset
)
{
    return gmpsp_setViBoolean (instrSession,
                               "SSET",
                               sset,
                               VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Sset
 * Purpose:  This function gets Signal Set (SSet) functionallity in the 
 *           instrument.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_GetSset(
    ViSession   instrSession,
    ViBoolean*  sset
)
{
    return gmpsp_getViBoolean (instrSession,
                               "SSET?",
                               sset,
                               VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set Analog Interface State
 * Purpose:  This function sets analog interface state.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_SetAnalogInterfaceState(
    ViSession   instrSession,
    ViInt32 uext,
    ViInt32 iext
)
{
    ViChar  buffer[BUFFER_SIZE] = "";
    
    CHECKPARAM (gmpsp_invalidViInt32Range (uext, 0, 2), 2);
    CHECKPARAM (gmpsp_invalidViInt32Range (iext, 0, 2), 3);
    
    sprintf (buffer, "ANALOG_IN %s,%s", analogArr[uext], analogArr[iext]);

    return gmpsp_sendCmd (instrSession, buffer);
}

/*****************************************************************************
 * Function: Get Analog Interface State
 * Purpose:  This function queries analog interface state.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_GetAnalogInterfaceState(
    ViSession   instrSession,
    ViInt32*    uext,
    ViInt32*    iext
)
{
    ViStatus gmpsp_status = 0;
    ViChar   buffer[BUFFER_SIZE];
    ViChar   *p2buf;
    
    CHECKERR (gmpsp_getViString (instrSession, "ANALOG_IN?", buffer, BUFFER_SIZE, 0));

    p2buf = strtok (buffer, ",");
    while (*p2buf == ' ') p2buf++;
    
    CHECKERR (gmpsp_getIndexFromString (p2buf, analogArr, uext));
    
    p2buf = strtok (NULL, ",");
    while (*p2buf == ' ') p2buf++;
    CHECKERR (gmpsp_getIndexFromString (p2buf, analogArr, iext));
    
    return gmpsp_status;
}

/*****************************************************************************
 * Function: Configure MinMax
 * Purpose:  This function configures U I Min Max measurement mode.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_ConfigureMinMax(
    ViSession   instrSession,
    ViBoolean   minmaxState,
    ViReal64    minVoltage,
    ViReal64    maxVoltage,
    ViReal64    minCurrent,
    ViReal64    maxCurrent
)
{
    ViStatus gmpsp_status = 0;
    
    gmpsp_status = gmpsp_SetMinMaxValues (instrSession, minVoltage, maxVoltage, minCurrent, maxCurrent);
    REMAP (2,3);
    REMAP (3,4);
    REMAP (4,5);
    REMAP (5,6);
    return gmpsp_SetMinMaxState (instrSession, minmaxState);
}

/*****************************************************************************
 * Function: Set MinMax State
 * Purpose:  This function selects U I Min Max mode.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_SetMinMaxState(
    ViSession   instrSession,
    ViBoolean   minmaxState
)
{
    return gmpsp_setViBoolean (instrSession,
                               "MINMAX",
                               minmaxState,
                               VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get MinMax State
 * Purpose:  This function queries U I Min Max mode state.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_GetMinMaxState(
    ViSession   instrSession,
    ViBoolean*  minmaxState
)
{
    return gmpsp_getViBoolean (instrSession,
                               "MINMAX?",
                               minmaxState,
                               VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set MinMax Values
 * Purpose:  This function selects limits for U I Min Max measurement mode.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_SetMinMaxValues(
    ViSession   instrSession,
    ViReal64    minVoltage,
    ViReal64    maxVoltage,
    ViReal64    minCurrent,
    ViReal64    maxCurrent
)
{
    ViChar  buffer[BUFFER_SIZE] = "";
    
    CHECKPARAM (gmpsp_invalidViReal64Range (minVoltage, 0.0, 60.0), 2);
    CHECKPARAM (gmpsp_invalidViReal64Range (maxVoltage, 0.0, 60.0), 3);
    CHECKPARAM (gmpsp_invalidViReal64Range (minCurrent, 0.0, 180.0), 4);
    CHECKPARAM (gmpsp_invalidViReal64Range (maxCurrent, 0.0, 180.0), 5);
    
    sprintf (buffer, "UI_C_SET %f,%f,%f,%f", minVoltage, maxVoltage, minCurrent, maxCurrent);

    return gmpsp_sendCmd (instrSession, buffer);
}

/*****************************************************************************
 * Function: Get MinMax Values
 * Purpose:  This function queries limits for U I Min Max measurement mode.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_GetMinMaxValues(
    ViSession   instrSession,
    ViReal64*   minVoltage,
    ViReal64*   maxVoltage,
    ViReal64*   minCurrent,
    ViReal64*   maxCurrent
)
{
    ViStatus gmpsp_status = 0;
    ViChar   buffer[BUFFER_SIZE];
    ViChar   *p2buf;
    
    CHECKERR (gmpsp_getViString (instrSession, "UI_C_SET?", buffer, BUFFER_SIZE, 0));

    p2buf = strtok (buffer, ",");
    *minVoltage = atof (p2buf);
    p2buf = strtok (NULL, ",");
    *maxVoltage = atof (p2buf);
    
    p2buf = strtok (NULL, ",");
    *minCurrent = atof (p2buf);
    p2buf = strtok (NULL, ",");
    *maxCurrent = atof (p2buf);
    
    return gmpsp_status;
}

/*****************************************************************************
 * Function: Set Power-On Mode
 * Purpose:  This function defines the status of the device setting after 
 *           power on.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_SetPwrOnMode(
    ViSession   instrSession,
    ViInt32 powerOnMode
)
{
    return gmpsp_setEnum (instrSession,
                          "POWER_ON",
                          powerOnMode,
                          powerOnArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Power-On Mode
 * Purpose:  This function queries the status of the device setting after 
 *           power on.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_GetPwrOnMode(
    ViSession   instrSession,
    ViInt32*    powerOnMode
)
{
    return gmpsp_getEnum (instrSession,
                          "POWER_ON?",
                          powerOnMode,
                          powerOnArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set Display Function
 * Purpose:  This function defines function of display A and display B.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_SetDisplayFunction(
    ViSession   instrSession,
    ViInt32 displayA,
    ViInt32 displayB
)
{
    ViChar  buffer[BUFFER_SIZE] = "";
    
    CHECKPARAM (gmpsp_invalidViInt32Range (displayA, 0, 3), 2);
    CHECKPARAM (gmpsp_invalidViInt32Range (displayB, 0, 3), 3);
    
    sprintf (buffer, "DISPLAY %s,%s", displayArr[displayA], displayBArr[displayB]);

    return gmpsp_sendCmd (instrSession, buffer);
}

/*****************************************************************************
 * Function: Get Display Function
 * Purpose:  This function queries function of display A and display B.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_GetDisplayFunction(
    ViSession   instrSession,
    ViInt32*    displayA,
    ViInt32*    displayB
)
{
    ViStatus gmpsp_status = 0;
    ViChar   buffer[BUFFER_SIZE];
    ViChar   *p2buf;
    
    CHECKERR (gmpsp_getViString (instrSession, "DISPLAY?", buffer, BUFFER_SIZE, 0));

    p2buf = strtok (buffer, ",");
    while (*p2buf == ' ') p2buf++;
    CHECKERR (gmpsp_getIndexFromString (p2buf, displayArr, displayA));
    
    p2buf = strtok (NULL, ",");
    while (*p2buf == ' ') p2buf++;
    CHECKERR (gmpsp_getIndexFromString (p2buf, displayBArr, displayB));
    
    return gmpsp_status;
}

/*****************************************************************************
 * Function: Set Measurement Filter
 * Purpose:  This function defines measurement low pass filter characteristic.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_SetMeasurementFilter(
    ViSession   instrSession,
    ViInt32 filter
)
{
    return gmpsp_setViInt32 (instrSession,
                             "MEAS_LPF",
                             filter,
                             1, 4,
                             VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Measurement Filter
 * Purpose:  This function queries measurement low pass filter characteristic.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_GetMeasurementFilter(
    ViSession   instrSession,
    ViInt32*    filter
)
{
    return gmpsp_getViInt32 (instrSession,
                             "MEAS_LPF?",
                             filter,
                             VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set Internal Sink State
 * Purpose:  This function switches on/off internal sink.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_SetInternalSinkState(
    ViSession   instrSession,
    ViBoolean   internalSinkState
)
{
    return gmpsp_setViBoolean (instrSession,
                               "SINK",
                               internalSinkState,
                               VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Internal Sink State
 * Purpose:  This function queries internal sink state.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_GetInternalSinkState(
    ViSession   instrSession,
    ViBoolean*  internalSinkState
)
{
    return gmpsp_getViBoolean (instrSession,
                               "SINK?",
                               internalSinkState,
                               VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set Constant Current Regulation
 * Purpose:  This function defines constant current regulation parameters.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_SetConstantCurrentRegulation(
    ViSession   instrSession,
    ViInt32 regulation
)
{
    return gmpsp_setEnum (instrSession,
                          "C_DYN",
                          regulation,
                          regulationArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Constant Current Regulation
 * Purpose:  This function queries constant current regulation parameters.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_GetConstantCurrentRegulation(
    ViSession   instrSession,
    ViInt32*    regulation
)
{
    return gmpsp_getEnum (instrSession,
                          "C_DYN?",
                          regulation,
                          regulationArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Configure Sequence Settings
 * Purpose:  This function configures sequence settings.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_ConfigureSeqSettings(
    ViSession   instrSession,
    ViInt32 startIndex,
    ViInt32 stopIndex,
    ViInt32 sequenceRepetition,
    ViReal64    defaultTime
)
{
    ViStatus gmpsp_status = 0;
    
    gmpsp_status = gmpsp_SetSeqRepetition (instrSession, sequenceRepetition);
    REMAP (2,4);
    
    gmpsp_status = gmpsp_SetSeqDefaultTime (instrSession, defaultTime);
    REMAP (2,5);

    return gmpsp_SetSeqRange (instrSession, startIndex, stopIndex);
}

/*****************************************************************************
 * Function: Configure Sequence Step
 * Purpose:  This function configures selected sequence step.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_ConfigureSeqStep(
    ViSession   instrSession,
    ViInt32 storageNumber,
    ViReal64    setVoltage,
    ViReal64    setCurrent,
    ViReal64    setTime,
    ViInt32 setFunction
)
{
    ViChar  buffer[BUFFER_SIZE] = "";
    
    CHECKPARAM (gmpsp_invalidViInt32Range (storageNumber, 1, 4096), 2);
    CHECKPARAM (gmpsp_invalidViReal64Range (setVoltage, 0.0, 60.0), 3);
    CHECKPARAM (gmpsp_invalidViReal64Range (setCurrent, 0.0, 180.0), 4);
    CHECKPARAM (gmpsp_invalidViReal64Range (setTime, 0.001, 65.535), 5);
    CHECKPARAM (gmpsp_invalidViInt32Range (setFunction, 0, 23), 6);
    
    sprintf (buffer, "STORE %ld,%f,%f,%f,%s", storageNumber, setVoltage, setCurrent, setTime, funcArr[setFunction]);

    return gmpsp_sendCmd (instrSession, buffer);
}

/*****************************************************************************
 * Function: Set Sequence Range
 * Purpose:  This function defines Sequence Range.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_SetSeqRange(
    ViSession   instrSession,
    ViInt32 startIndex,
    ViInt32 stopIndex
)
{
    ViChar  buffer[BUFFER_SIZE] = "";
    
    CHECKPARAM (gmpsp_invalidViInt32Range (startIndex, 1, 4096), 2);
    CHECKPARAM (gmpsp_invalidViInt32Range (stopIndex, 1, 4096), 3);
    
    sprintf (buffer, "START_STOP %ld,%ld", startIndex, stopIndex);

    return gmpsp_sendCmd (instrSession, buffer);
}

/*****************************************************************************
 * Function: Get Sequence Range
 * Purpose:  This function queries Sequence Range.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_GetSeqRange(
    ViSession   instrSession,
    ViInt32*    startIndex,
    ViInt32*    stopIndex
)
{
    ViStatus gmpsp_status = 0;
    ViChar   buffer[BUFFER_SIZE];
    ViChar   *p2buf;
    
    CHECKERR (gmpsp_getViString (instrSession, "START_STOP?", buffer, BUFFER_SIZE, 0));
    p2buf = strtok (buffer, ",");
    *startIndex = atol (p2buf);
    p2buf = strtok (NULL, ",");
    *stopIndex = atol (p2buf);
    
    return gmpsp_status;
}

/*****************************************************************************
 * Function: Set Sequence Repetition
 * Purpose:  This function defines Sequence Repetition.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_SetSeqRepetition(
    ViSession   instrSession,
    ViInt32 sequenceRepetition
)
{
    return gmpsp_setViInt32 (instrSession,
                             "REPETITION",
                             sequenceRepetition,
                             0, 255,
                             VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Sequence Repetition
 * Purpose:  This function queries Sequence Repetition.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_GetSeqRepetition(
    ViSession   instrSession,
    ViInt32*    sequenceRepetition
)
{
    return gmpsp_getViInt32 (instrSession,
                             "REPETITION?",
                             sequenceRepetition,
                             VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set Sequence Default Time
 * Purpose:  This function defines Default Sequence Time.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_SetSeqDefaultTime(
    ViSession   instrSession,
    ViReal64    defaultTime
)
{
    return gmpsp_setViReal64 (instrSession,
                              "TDEF",
                              defaultTime,
                              0.001, 65.535,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Sequence Default Time
 * Purpose:  This function queries Default Sequence Time.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_GetSeqDefaultTime(
    ViSession   instrSession,
    ViReal64*   defaultTime
)
{
    return gmpsp_getViReal64 (instrSession,
                              "TDEF?",
                              defaultTime,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Sequence Control
 * Purpose:  This function operates the sequence trigger.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_actSendSeqTrigger(
    ViSession   instrSession,
    ViInt32 sequenceControl
)
{
    return gmpsp_setEnum (instrSession,
                          "SEQUENCE",
                          sequenceControl,
                          triggerArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Sequence Status
 * Purpose:  This function operates the sequence trigger.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_actSeqStatus (ViSession instrSession,
                                      ViInt32 *sequenceStatus,
                                      ViInt32 *sequenceNumber,
                                      ViInt32 *repetition, ViInt32 *step)
{
    ViStatus gmpsp_status = 0;
    ViChar   buffer[BUFFER_SIZE];
    ViChar   *p2buf;
    
    CHECKERR (gmpsp_getViString (instrSession, "SEQUENCE?", buffer, BUFFER_SIZE, 0));
    p2buf = strtok (buffer, ",");
    
    if (sequenceStatus != NULL)
    {
        CHECKERR (gmpsp_getIndexFromString (p2buf, triggerStatusArr, sequenceStatus));
    }
    
    p2buf = strtok (NULL, ",");
    if (sequenceNumber != NULL)
    {
        *sequenceNumber = atol (p2buf);
    }
    
    p2buf = strtok (NULL, ",");
    if (repetition != NULL)
    {
        *repetition = atol (p2buf);
    }

    p2buf = strtok (NULL, ",");
    if (step != NULL)
    {
        *step = atol (p2buf);
    }
    return gmpsp_status;
}

/*****************************************************************************
 * Function: Reset MinMax
 * Purpose:  This function resets values in MinMax memory.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_resetMinMax(
    ViSession   instrSession
)
{
    return gmpsp_sendCmd (instrSession, "MINMAX RST");
}

/*****************************************************************************
 * Function: Protection Status
 * Purpose:  This function queries status of overvoltage and overcurrent 
 *           protection.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_actProtectionStatus(
    ViSession   instrSession,
    ViBoolean*  ovpStatus,
    ViBoolean*  ocpStatus
)
{
    ViStatus gmpsp_status = 0;
    ViInt32  reg_val = 0;
    
    CHECKERR (viQueryf (instrSession, "CRA?", "%ld", &reg_val));
    
    if (ovpStatus != NULL)
        *ovpStatus = ((reg_val & 16) != 0) ? 1 : 0;
    if (ocpStatus != NULL)
        *ocpStatus = ((reg_val & 8) != 0) ? 1 : 0;
    return gmpsp_status;
}

/*****************************************************************************
 * Function: Fetch Min Max Measurement
 * Purpose:  This function returns the limit values of the output current and 
 *           voltage measured and saved in MINMAX memory during MINMAX ON.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_fetchMinMax(
    ViSession   instrSession,
    ViReal64*   minVoltage,
    ViReal64*   maxVoltage,
    ViReal64*   minCurrent,
    ViReal64*   maxCurrent
)
{
    ViStatus gmpsp_status = 0;
    
    if (minVoltage != NULL)
    {
        CHECKERR (gmpsp_getViReal64 (instrSession, "UMIN?", minVoltage, VI_ERROR_PARAMETER2));
    }
    
    if (maxVoltage != NULL)
    {
        CHECKERR (gmpsp_getViReal64 (instrSession, "UMAX?", maxVoltage, VI_ERROR_PARAMETER3));
    }
        
    if (minCurrent != NULL)
    {
        CHECKERR (gmpsp_getViReal64 (instrSession, "IMIN?", minCurrent, VI_ERROR_PARAMETER4));
    }
    if (maxCurrent != NULL)
    {
        CHECKERR (gmpsp_getViReal64 (instrSession, "IMAX?", maxCurrent, VI_ERROR_PARAMETER5));
    }
    
    return gmpsp_status;
}

/*****************************************************************************
 * Function: Meas Output Voltage
 * Purpose:  This function queries the actual measured value of the output 
 *           voltage.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_measOutVolt(
    ViSession   instrSession,
    ViReal64*   outputVoltage
)
{
    return gmpsp_getViReal64 (instrSession,
                              "UOUT?",
                              outputVoltage,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Meas Output Current
 * Purpose:  This function queries the actual measured value of the output 
 *           current.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_measOutCurr(
    ViSession   instrSession,
    ViReal64*   outputCurrent
)
{
    return gmpsp_getViReal64 (instrSession,
                              "IOUT?",
                              outputCurrent,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Meas Output Power
 * Purpose:  This function queries the actually measured power on the output.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_measOutPower(
    ViSession   instrSession,
    ViReal64*   outputPower
)
{
    return gmpsp_getViReal64 (instrSession,
                              "POUT?",
                              outputPower,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: State Checking
 * Purpose:  This function switches state checking of the instrument (reading 
 *           of the Standard Event Register and checking it for error) status 
 *           subsystem. Driver functions are using state checking which is by 
 *           default enabled.
 *           
 *           Note:
 *           
 *           (1) In debug mode enable state checking.
 *           
 *           (2) For better bus throughput and instruments performance disable 
 *           state checking.
 *           
 *           (3) When state checking is disabled driver does not check if 
 *           correct instrument model or option is used with each of the 
 *           functions. This might cause unexpected behaviour of the 
 *           instrument.
 *           
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_errorCheckState(
    ViSession   instrSession,
    ViBoolean   stateChecking
)
{
    ViStatus    gmpsp_status = VI_SUCCESS;
    gmpsp_instrPtr instrPtr = NULL;

    CHECKERR (viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr));

    CHECKNULL (instrPtr, 1);
    
    instrPtr -> errorChecking = stateChecking;

    return gmpsp_status;
}

/*****************************************************************************
 * Function: Set Status Register
 * Purpose:  This function sets device dependend event enable registers.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_SetStatusRegister(
    ViSession   instrSession,
    ViInt32 reg,
    ViInt32 enable
)
{
    CHECKPARAM (gmpsp_invalidViInt32Range (reg, 0, 2), 2);
    
    return gmpsp_setViInt32 (instrSession,
                             registerEnableArr[reg],
                             enable,
                             0, 255,
                             VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Get Status Register
 * Purpose:  This function contains the commands for the querying status 
 *           reporting system.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_getStatusRegister(
    ViSession   instrSession,
    ViInt32 reg,
    ViInt32*    registerValue
)
{
    ViChar  buffer[BUFFER_SIZE] = "";
    
    CHECKPARAM (gmpsp_invalidViInt32Range (reg, 0, 7), 2);
    
    sprintf (buffer, "%s?", registerEnableArr[reg]);
    
    return gmpsp_getViInt32 (instrSession,
                             buffer,
                             registerValue,
                             VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Save Setup
 * Purpose:  This function stores present setup as the user defined setup in 
 *           storages.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_SaveSetup(
    ViSession   instrSession,
    ViInt32 userSetup
)
{
    return gmpsp_setViInt32 (instrSession,
                             "*SAV",
                             userSetup,
                             1, 12,
                             VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Recall Setup
 * Purpose:  This function recalls the setup.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_RecallSetup(
    ViSession   instrSession,
    ViInt32 setupNumber
)
{
    return gmpsp_setViInt32 (instrSession,
                             "*RCL",
                             setupNumber,
                             1, 12,
                             VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Write To Instrument
 * Purpose:  This function writes commands and queries to the instrument to 
 *           modify parameters and query device settings.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_writeInstrData(
    ViSession   instrSession,
    ViChar   writeBuffer[]
)
{
    ViStatus gmpsp_status = VI_SUCCESS;

    CHECKERR (viPrintf (instrSession, "%s\n", writeBuffer));

    return gmpsp_status;
}

/*****************************************************************************
 * Function: Read Instrument Data
 * Purpose:  This function reads data from the instrument's output buffer and 
 *           returns it to the specified variable in memory.
 *           
 *           Note:
 *           1) Because the instrument may return both numeric and text data 
 *           in response to queries this function returns the data in string 
 *           format.
 *           2) If valid data is not available at the instrument's output 
 *           buffer when this function is called the instrument will hang up 
 *           and the function will not return until it times out.  If the 
 *           time-out is disabled this function will hang indefinitely and it 
 *           may be necessary to reboot the computer to break out.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_readInstrData(
    ViSession   instrSession,
    ViInt32 numberBytesToRead,
    ViChar    readBuffer[],
    ViInt32*    numbytesRead
)
{
    ViStatus    gmpsp_status = VI_SUCCESS;
    ViUInt32    tmpCnt = 0;

    if (numbytesRead != NULL)
        *numbytesRead = 0L;

    CHECKERR (viRead (instrSession, (ViBuf)readBuffer, (ViUInt32) numberBytesToRead, &tmpCnt));

    if (numbytesRead)
        *numbytesRead = (ViInt32)tmpCnt;

    return gmpsp_status;
}

/*****************************************************************************
 * Function: Reset
 * Purpose:  This function resets the instrument and clears all event 
 *           registers and error queues.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_reset(
    ViSession   instrSession
)
{
    ViStatus gmpsp_status = VI_SUCCESS;
    
    CHECKERR (gmpsp_sendCmd (instrSession, "*RST"));
    (void) gmpsp_delay (instrSession, 0.1); 
    return gmpsp_status;
}

/*****************************************************************************
 * Function: Self-Test
 * Purpose:  This function returns result of self-test function.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_self_test(
    ViSession   instrSession,
    ViInt16*    selfTestResult,
    ViChar      selfTestMessage[]
)
{
    ViStatus gmpsp_status = 0;
    ViInt32 value = 0;

    CHECKERR (gmpsp_getViInt32 (instrSession,
                                "*TST?",
                                &value,
                                VI_ERROR_PARAMETER2));
    if (selfTestResult != NULL)
        *selfTestResult = (ViUInt16) value;
    
    return gmpsp_status;
}

/*****************************************************************************
 * Function: Error Query
 * Purpose:  This function reads an error code from the instrument's error 
 *           queue.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_error_query(
    ViSession   instrSession,
    ViInt32*    errorCode,
    ViChar      errorMessage[]
)
{
    ViStatus gmpsp_status = 0;
    ViChar   buffer[BUFFER_SIZE];
    
    CHECKERR (gmpsp_getViString (instrSession, "ERROR?", buffer, BUFFER_SIZE, 0));
    if (errorCode != NULL)
        *errorCode = atol (buffer);
    
    return gmpsp_status;
}

/*****************************************************************************
 * Function: Error Message
 * Purpose:  This function takes the Status Code returned by the instrument 
 *           driver functions, interprets it and returns it as a user readable 
 *           string. 
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_error_message(
    ViSession   instrSession,
    ViStatus    statusCode,
    ViChar      errMessage[]
)
{
    ViStatus gmpsp_status = VI_SUCCESS;
    ViInt16 i;

    static gmpsp_tStringValPair statusDescArray[] = {
        {VI_ERROR_PARAMETER1,                   "ERROR: Parameter 1 out of range"},
        {VI_ERROR_PARAMETER2,                   "ERROR: Parameter 2 out of range"},
        {VI_ERROR_PARAMETER3,                   "ERROR: Parameter 3 out of range"},
        {VI_ERROR_PARAMETER4,                   "ERROR: Parameter 4 out of range"},
        {VI_ERROR_PARAMETER5,                   "ERROR: Parameter 5 out of range"},
        {VI_ERROR_PARAMETER6,                   "ERROR: Parameter 6 out of range"},
        {VI_ERROR_PARAMETER7,                   "ERROR: Parameter 7 out of range"},
        {VI_ERROR_PARAMETER8,                   "ERROR: Parameter 8 out of range"},
        {VI_ERROR_FAIL_ID_QUERY,                "ERROR: Identification query failed"},
        {VI_ERROR_INV_RESPONSE,                 "ERROR: Interpreting instrument response"},
        {VI_ERROR_INSTR_INTERPRETING_RESPONSE,  "ERROR: Interpreting the instrument's response"},

        /* --- Instrument specific error codes --- */

        {GMPSP_ERROR_INSTRUMENT_ERROR,          "ERROR: Instrument status error"},

        /* --- Instrument specific warnings --- */

        {GMPSP_WARNING_SMALL_BUFFER,            "WARNING: Buffer too small"},

        {VI_NULL, VI_NULL}
    };

    gmpsp_status = viStatusDesc (instrSession, statusCode, errMessage);
    if (gmpsp_status == VI_WARN_UNKNOWN_STATUS)
        {
        for (i = 0; statusDescArray[i].stringName; i++)
            {
            if (statusDescArray[i].stringVal == statusCode)
                {
                sprintf (errMessage, "%s", statusDescArray[i].stringName);
                return (VI_SUCCESS);
                }
            }
        sprintf (errMessage, "Unknown Error 0x%X", (unsigned int) statusCode);
        return (VI_WARN_UNKNOWN_STATUS);
        }

    gmpsp_status = VI_SUCCESS;
    return gmpsp_status;
}

/*****************************************************************************
 * Function: Revision Query
 * Purpose:  This function returns the revision numbers of the instrument 
 *           driver and instrument firmware, and tells the user with which  
 *           instrument firmware this revision of the driver is compatible. 
 *           
 *           
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_revision_query(
    ViSession   instrSession,
    ViChar      instrumentDriverRevision[],
    ViChar      firmwareRevision[]
)
{
    ViStatus gmpsp_status = VI_SUCCESS;

    CHECKERR (viPrintf (instrSession, "*IDN?\n"));
    CHECKERR (viScanf (instrSession, "%*[^ ] %*[^,],%*[^,],%*[^,],%[^\r\n]", firmwareRevision));

    sprintf (instrumentDriverRevision, "%s", GMPSP_REVISION);

    return gmpsp_status;
}

/*****************************************************************************
 * Function: Close
 * Purpose:  This function performs the following operations:
 *           viClose (instrSession) and viClose (rmSession).
 *           
 *           Notes:
 *           
 *           (1) The instrument must be reinitialized to use it again.
 *****************************************************************************/
ViStatus _VI_FUNC gmpsp_close(
    ViSession   instrSession
)
{
    gmpsp_instrPtr instrPtr = NULL;
    ViSession rmSession = 0;
    ViStatus gmpsp_status = VI_SUCCESS;

    CHECKERR( viGetAttribute (instrSession, VI_ATTR_RM_SESSION, &rmSession));
    CHECKERR( viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr));

    if (instrPtr != NULL)
        free (instrPtr);

    gmpsp_status = viClose (instrSession);
    (void) viClose (rmSession);

    return gmpsp_status;
}

/*===========================================================================*/
/*----------- INSERT INSTRUMENT-DEPENDENT UTILITY ROUTINES HERE -------------*/
/*===========================================================================*/

/*===========================================================================*/
/* Function: Default Instrument Setup
/* Purpose:  This function sends a default setup to the instrument.  This
/*           function is called by the gmpsp_reset operation and by the
/*           gmpsp_init function if the reset option has not been
/*           selected.  This function is useful for configuring any
/*           instrument settings that are required by the rest of the
/*           instrument driver functions such as turning headers ON or OFF
/*           or using the long or short form for commands, queries, and data.*/
/*===========================================================================*/
ViStatus gmpsp_defaultInstrSetup (ViSession instrSession)
{
    ViStatus            gmpsp_status = VI_SUCCESS;
    gmpsp_instrPtr      instrPtr = NULL;

    CHECKERR (viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr));
    
    CHECKNULL (instrPtr, 1);
    
    /* Determine if the structure has been initialized for the current VISA  */
    /* Session and malloc if it has not.                                     */
    CHECKERR (viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr));

    instrPtr -> errorChecking = VI_TRUE;
    strcpy (instrPtr -> instrDriverRevision, GMPSP_REVISION);

    return gmpsp_status;
}

/*****************************************************************************
 * Function: Initialize Clean Up
 * Purpose:  This function is used only by the gmpsp_init function.  When
 *           an error is detected this function is called to close the
 *           open resource manager and instrument object sessions and to
 *           set the instrSession that is returned from gmpsp_init to
 *           VI_NULL.
 *****************************************************************************/
ViStatus gmpsp_initCleanUp (ViSession openRMSession,
                             ViPSession openInstrSession,
                             ViStatus currentStatus)
{
    gmpsp_instrPtr instrPtr = NULL;

    if (viGetAttribute (*openInstrSession, VI_ATTR_USER_DATA, &instrPtr) >= 0)
        if (instrPtr != NULL)
            free (instrPtr);

    (void) viClose (*openInstrSession);
    (void) viClose (openRMSession);
    *openInstrSession = VI_NULL;

    return currentStatus;
}
